@@ -27,6 +27,8 @@ from beaker.util import b64decode, b64encode, Set
 
 __all__ = ['SignedCookie', 'Session']
 
+getpid = hasattr(os, 'getpid') and os.getpid or (lambda : '')
+
 class SignedCookie(Cookie.BaseCookie):
     """Extends python cookie to give digital signature support"""
     def __init__(self, secret, input=None):
@@ -76,6 +78,7 @@ class Session(dict):
         self.secret = secret
         self.secure = secure
         self.id = id
+        self.accessed_dict = {}
             
         if self.use_cookies:
             cookieheader = request.get('cookie', '')
@@ -105,7 +108,7 @@ class Session(dict):
     def _create_id(self):
         self.id = md5(
             md5("%f%s%f%s" % (time.time(), id({}), random.random(),
-                              os.getpid()) ).hexdigest(), 
+                              getpid())).hexdigest(), 
         ).hexdigest()
         self.is_new = True
         if self.use_cookies:
@@ -134,18 +137,6 @@ class Session(dict):
         return self['_creation_time']
     created = property(created)
 
-    def delete(self):
-        """Deletes the session from the persistent storage, and sends
-        an expired cookie out"""
-        if self.use_cookies:
-            self._delete_cookie()
-        if hasattr(self, 'namespace'):
-            self.namespace.acquire_write_lock()
-            try:
-                self.namespace.remove()
-            finally:
-                self.namespace.release_write_lock()
-
     def _delete_cookie(self):
         self.request['set_cookie'] = True
         self.cookie[self.key] = self.id
@@ -160,54 +151,70 @@ class Session(dict):
         self.request['cookie_out'] = self.cookie[self.key].output(header='')
         self.request['set_cookie'] = True
 
+    def delete(self):
+        """Deletes the session from the persistent storage, and sends
+        an expired cookie out"""
+        if self.use_cookies:
+            self._delete_cookie()
+        self.clear()
+
     def invalidate(self):
         """Invalidates this session, creates a new session id, returns
         to the is_new state"""
-        if hasattr(self, 'namespace'):
-            namespace = self.namespace
-            namespace.acquire_write_lock()
-            try:
-                namespace.remove()
-            finally:
-                namespace.release_write_lock()
-            
+        self.clear()
         self.was_invalidated = True
         self._create_id()
         self.load()
- 
+
+    def _session_dict_from_namespace(self):
+        now = time.time()
+        try:
+            session_data = self.namespace['session']
+            session_data['_accessed_time'] = now
+        except KeyError:
+            session_data = {
+                '_creation_time':now,
+                '_accessed_time':now
+            }
+            self.is_new = True
+        self.accessed = session_data['_accessed_time']
+        return session_data
+    
+    def _persist_session_dict(self, session_dict):
+        if not session_dict and 'session' in self.namespace:
+            del self.namespace['session']
+        else:
+            self.namespace['session'] = session_dict
+        
     def load(self):
         "Loads the data from this session from persistent storage"
-        self.namespace = namespace = self.namespace_class(self.id,
+        self.namespace = self.namespace_class(self.id,
             data_dir=self.data_dir, digest_filenames=False,
             **self.namespace_args)
-        
+        now = time.time()
         self.request['set_cookie'] = True
         
-        namespace.acquire_write_lock()
+        self.namespace.acquire_read_lock()
         try:
             self.clear()
-            now = time.time()
-            
-            if not '_creation_time' in namespace:
-                namespace['_creation_time'] = now
-                self.is_new = True
-            try:
-                self.accessed = namespace['_accessed_time']
-                namespace['_accessed_time'] = now
-            except KeyError:
-                namespace['_accessed_time'] = self.accessed = now
+            session_data = self._session_dict_from_namespace()
             
             if self.timeout is not None and now - self.accessed > self.timeout:
                 self.invalidate()
             else:
-                for k in namespace.keys():
-                    self[k] = namespace[k]
-        
+                self.update(session_data)
+            self.accessed_dict = session_data.copy()
         finally:
-            namespace.release_write_lock()
+            self.namespace.release_read_lock()
     
-    def save(self):
-        "Saves the data for this session to persistent storage"
+    def save(self, accessed_only=False):
+        """Saves the data for this session to persistent storage
+        
+        If accessed_only is True, then only the original data loaded
+        at the beginning of the request will be saved, with the updated
+        last accessed time.
+        
+        """
         if not hasattr(self, 'namespace'):
             self.namespace = self.namespace_class(
                                     self.id, 
@@ -217,24 +224,25 @@ class Session(dict):
 
         self.namespace.acquire_write_lock()
         try:
-            if not '_creation_time' in self.namespace:
-                self.is_new = True
-            
-            keys = self.keys()
-            for k in Set(self.namespace.keys()).difference(keys):
-                del self.namespace[k]
-                    
-            for k in keys:
-                self.namespace[k] = self[k]
-            
-            now = time.time()
-            self.namespace['_accessed_time'] = self['_accessed_time'] = now
-            self.namespace['_creation_time'] = self['_creation_time'] = now
+            if accessed_only:
+                data = dict(self.accessed_dict.items())
+            else:
+                data = dict(self.items())
+            self._persist_session_dict(data)
         finally:
             self.namespace.release_write_lock()
         if self.is_new:
             self.request['set_cookie'] = True
     
+    def revert(self):
+        """Revert the session to its original state from its first
+        access in the request"""
+        self.clear()
+        self.update(self.accessed_dict)
+    
+    # TODO: I think both these methods should be removed.  They're from
+    # the original mod_python code i was ripping off but they really
+    # have no use here.
     def lock(self):
         """Locks this session against other processes/threads.  This is
         automatic when load/save is called.
@@ -324,6 +332,7 @@ class CookieSession(Session):
             if self.timeout is not None and time.time() - \
                self['_accessed_time'] > self.timeout:
                 self.clear()
+            self.accessed_dict = self.copy()
             self._create_cookie()
     
     def created(self):
@@ -364,20 +373,45 @@ class CookieSession(Session):
     
     def _make_id(self):
         return md5(md5(
-            "%f%s%f%d" % (time.time(), id({}), random.random(), os.getpid())
+            "%f%s%f%d" % (time.time(), id({}), random.random(), getpid())
             ).hexdigest()
         ).hexdigest()
     
-    def save(self):
+    def save(self, accessed_only=False):
         """Saves the data for this session to persistent storage"""
+        if accessed_only:
+            self.clear()
+            self.update(self.accessed_dict)
         self._create_cookie()
     
+    def expire(self):
+        """Delete the 'expires' attribute on this Session, if any."""
+        
+        self.pop('_expires', None)
+        
     def _create_cookie(self):
         if '_creation_time' not in self:
             self['_creation_time'] = time.time()
         if '_id' not in self:
             self['_id'] = self._make_id()
         self['_accessed_time'] = time.time()
+        
+        if self.cookie_expires is not True:
+            if self.cookie_expires is False:
+                expires = datetime.fromtimestamp( 0x7FFFFFFF )
+            elif isinstance(self.cookie_expires, timedelta):
+                expires = datetime.today() + self.cookie_expires
+            elif isinstance(self.cookie_expires, datetime):
+                expires = self.cookie_expires
+            else:
+                raise ValueError("Invalid argument for cookie_expires: %s"
+                                 % repr(self.cookie_expires))
+            self['_expires'] = expires
+        elif '_expires' in self:
+            expires = self['_expires']
+        else:
+            expires = None
+
         val = self._encrypt_data()
         if len(val) > 4064:
             raise BeakerException("Cookie value is too long to store")
@@ -389,16 +423,8 @@ class CookieSession(Session):
             self.cookie[self.key]['secure'] = True
         
         self.cookie[self.key]['path'] = '/'
-        if self.cookie_expires is not True:
-            if self.cookie_expires is False:
-                expires = datetime.fromtimestamp( 0x7FFFFFFF )
-            elif isinstance(self.cookie_expires, timedelta):
-                expires = datetime.today() + self.cookie_expires
-            elif isinstance(self.cookie_expires, datetime):
-                expires = self.cookie_expires
-            else:
-                raise ValueError("Invalid argument for cookie_expires: %s"
-                                 % repr(self.cookie_expires))
+        
+        if expires:
             self.cookie[self.key]['expires'] = \
                 expires.strftime("%a, %d-%b-%Y %H:%M:%S GMT" )
         self.request['cookie_out'] = self.cookie[self.key].output(header='')
@@ -474,6 +500,31 @@ class SessionObject(object):
         params = self.__dict__['_params']
         session = Session({}, use_cookies=False, id=id, **params)
         if session.is_new:
-            session.namespace.remove()
             return None
         return session
+    
+    def save(self):
+        self.__dict__['_dirty'] = True
+    
+    def persist(self):
+        """Persist the session to the storage
+        
+        If its set to autosave, then the entire session will be saved
+        regardless of if save() has been called. Otherwise, just the
+        accessed time will be updated if save() was not called, or
+        the session will be saved if save() was called.
+        
+        """
+        if self.__dict__['_params'].get('auto'):
+            self._session().save()
+        else:
+            if self.__dict__['_dirty']:
+                self._session().save()
+            else:
+                self._session().save(accessed_only=True)
+    
+    def dirty(self):
+        return self.__dict__.get('_dirty', False)
+    
+    def accessed(self):
+        return self.__dict__['_sess'] is not None
